/**************************************************************************
Revision 1.0.0 of the freeLoader console program for release
Last update: May 23, 2011

Copyright (c) 2011, Hod Lipson and John Amend
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer.

 * Redistributions in binary form must reproduce the above copyright 
   notice, this list of conditions and the following disclaimer in the 
   documentation and/or other materials provided with the distribution.

 * Neither the name of the freeLoader nor the names of its contributors 
   may be used to endorse or promote products derived from this software 
   without specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
**************************************************************************/

#include <iostream>		//for I/O
#include <iomanip>		//for I/O manipulation
#include <cstdio>		//more inputs & outputs
#include <fstream>		//for reading/writing to files
#include <string>
#include <windows.h>	
#include "tic_toc.h"	//high performance timer using QueryPerformanceCounter
#include "ftd2xx.h"		//FTDI USB to Serial functions
#include "motor_funcs.h"//Motor specific functions
#include <cassert>
#include <conio.h>		//for e-stop escape
#include <sstream>		//converting double to string

#pragma comment(lib, "Winmm.lib")
#pragma comment(lib, "ftd2xx.lib")


int main()
{
	//start the timer
		init_timer();


	// Welcome splash
		std::cout << "\n*******************************************************************************";
		std::cout << "\n*                                                                             *";
		std::cout << "\n*   freeLoader control software, V 1.0.0 by John Amend - Cornell University   *";
		std::cout << "\n*                                                                             *";
		std::cout << "\n*      can control 1 crosshead at a time for 1 machine with 2 crossheads      *";
		std::cout << "\n*                                                                             *"; 
		std::cout << "\n*******************************************************************************\n";

	//Run a test or move the crossheads
		int ToDo;
		std::cout << "\nWhat do you want to do? 1) run a test, 2) move a crosshead (enter 1 or 2): ";
		std::cin >> ToDo;
		if(ToDo != 1 && ToDo != 2)
		{
			std::cerr <<"\nInvalid Entry!";
			Sleep(2000);
			return 0;
		}


	//Look for FTDI Devices
		//Get the number of devices currently connected and store this in numDevs
		FT_STATUS ftStatus; 
		DWORD numDevs;
		ftStatus = FT_ListDevices(&numDevs,NULL,FT_LIST_NUMBER_ONLY); 
		if (ftStatus == FT_OK)
			std::cout << "\nUSB to Serial devices detected: " << numDevs;
		else
		{
			std::cerr << "\nConnection to FT_ListDevices failed!"; 
			Sleep(2000);
			return 0;
		}
		if (numDevs==0)
		{
			std::cerr << "\n\nNo devices found. Exiting program.";
			Sleep(2000);
			return 0;
		}
		

	//List device serial numbers 
		DWORD devIndex;
		char Buffer [3][64]; // more than enough room, maximum 3 USB to serial devices
		for(int i=0;i<numDevs;i++) //for each device detected
		{
			devIndex = i;						
			ftStatus = FT_ListDevices((PVOID)devIndex,Buffer[i],FT_LIST_BY_INDEX|FT_OPEN_BY_SERIAL_NUMBER); 					
			if (ftStatus == FT_OK) 				
				std::cout << "\n Device " << i+1 << " serial number: " << Buffer[i];
			else 
			{
				std::cerr << "\nFT_ListDevices failed to get serial number!";			
				Sleep(2000);
			    return 0;
			}
		}

					
	//Identify UB2Dynamixel 
		int dyna_dev_num;
		char dyna_serial[64];
		char *DYNA_serial = dyna_serial;		
		std::cout << "\n\nEnter the device that matches the USB2Dynamixel serial number (ie 1, 2 or 3): ";
		std::cin >> dyna_dev_num;				
		DYNA_serial = Buffer[dyna_dev_num -1];
		if (dyna_dev_num < 1 || dyna_dev_num > numDevs)
		{
			std::cerr <<"\nInvalid Entry!";
			Sleep(2000);
			return 0;
		}
	

	//Open USB2Dynamixel
		FT_HANDLE ftHandleDYNA;
		ftStatus = FT_OpenEx(DYNA_serial,FT_OPEN_BY_SERIAL_NUMBER,&ftHandleDYNA); 		
		if (ftStatus != FT_OK)
		{
			std::cerr << "\n\nUSB2Dynamixel failed to open!";	
			Sleep(2000);
			return 0;
		}

	//Initialize the USB2Dynamixel
		DYNA_initialize(ftHandleDYNA);
		Sleep(50);


	//Moving crossheads
		if(ToDo == 2) //if the user wants to move the crossheads
		{
			int motorNumber;
			std::cout<<"\nWhich motor/crosshead do you want to move (1 or 2)?:";
			std::cin>>motorNumber;
			if(motorNumber!=1 && motorNumber!=2)
			{
				std::cerr <<"\nInvalid Entry!";
				Sleep(2000);
				return 0;
			}

			int moveDirection;
			std::cout << "\nWhat direction to move the crosshead (1 = up, 2 = down)?: ";
			std::cin >> moveDirection;
			if (moveDirection !=1 && moveDirection != 2)
			{
				std::cerr <<"\nInvalid Entry!";
				Sleep(2000);
				return 0;
			}

			int moveSpeed;
			std::cout << "\nHow fast (1 = fast, 2 = slow)?: ";
			std::cin >> moveSpeed;
			if (moveSpeed !=1 && moveSpeed != 2)
			{
				std::cerr <<"\nInvalid Entry!";
				Sleep(2000);
				return 0;
			}
			
			//Calculate motor speed from inputs
			int motorSpeed;
			if (moveDirection==1 && moveSpeed==1) // move up fast
				motorSpeed = 2023;
			else if (moveDirection==1 && moveSpeed==2) // move up slow
				motorSpeed = 1400;
			else if (moveDirection==2 && moveSpeed==1) // move down fast
				motorSpeed = 1000;
			else // move down slow
				motorSpeed = 400;


			std::cout<< "\n\n\n          PRESS ANY KEY TO STOP MOTOR\n\n\n";			
	

			//Spin the motor			
				motor_spin(ftHandleDYNA, motorNumber, motorSpeed);

			while(1)
			{
				//check for e-stop exit
				if( _kbhit() > 0 )
				{					
					break;	
				}
			}

			//shut off the motor
			motor_spin(ftHandleDYNA, motorNumber, 0);

			//Close USB2Dynamixel
			ftStatus = FT_Close(ftHandleDYNA); 	
			if (ftStatus != FT_OK) 		
			std::cerr << "\n\nUSB2Dynamixel failed to close";

			//display "press any key..." to close the console
			std::cout << "\n";
			system("pause"); 

			return 0;
		}





	//Running a test
		else //if the user wants to run a test
		{

			int motorNumber;
			std::cout<<"\nWhich motor/crosshead do you want to move (1 or 2)?:";
			std::cin>>motorNumber;
			if(motorNumber!=1 && motorNumber!=2)
			{
				std::cerr <<"\nInvalid Entry!";
				Sleep(2000);
				return 0;
			}

			//Get comport for load cell
				std::wstring comNum;
				std::wcout << L"\n\nEnter COM port for accompanying load cell (ex: COM20): ";
				std::wcin >> comNum;
				std::wstring comPrefix = L"\\\\.\\";
				std::wstring comID = comPrefix+comNum;


			//Open serial port at comID
				HANDLE hSerial;
				hSerial = CreateFileW(	comID.c_str(),
										GENERIC_READ | GENERIC_WRITE,
										0,
										0,
										OPEN_EXISTING,
										FILE_ATTRIBUTE_NORMAL,
										0);
				if(hSerial==INVALID_HANDLE_VALUE)
				{
					std::cerr << "\n\nError opening port!";	
					Sleep(2000);
					return 0;
				}

			//Set serial port parameters
				DCB dcbSerialParams = {0};
				dcbSerialParams.DCBlength=sizeof(dcbSerialParams);

				if (!GetCommState(hSerial, &dcbSerialParams)) 
				{
					std::cerr << "\n\nError getting port state!";
					Sleep(2000);
					return 0;
				}

				dcbSerialParams.BaudRate=CBR_9600;
				dcbSerialParams.ByteSize=8;
				dcbSerialParams.StopBits=ONESTOPBIT;
				dcbSerialParams.Parity=NOPARITY;

				if(!SetCommState(hSerial, &dcbSerialParams))
				{
					std::cerr << "\n\nError setting port state!";
					Sleep(2000);
					return 0;
				}

			//Setting timeouts
				COMMTIMEOUTS timeouts={0};
				timeouts.ReadIntervalTimeout=50;
				timeouts.ReadTotalTimeoutConstant=50;
				timeouts.ReadTotalTimeoutMultiplier=10;
				timeouts.WriteTotalTimeoutConstant=50;
				timeouts.WriteTotalTimeoutMultiplier=10;

				if(!SetCommTimeouts(hSerial, &timeouts))
				{
					std::cerr << "\n\nError setting timeouts!";
					Sleep(2000);
					return 0;
				}

			//Write a <CR> to the port.		
				DWORD dwBytesWrite = 0;
				if(!WriteFile(hSerial, "\r", 1, &dwBytesWrite, NULL))
				{
					std::cerr << "\n\n error occurred writing CR to port";		
					Sleep(2000);
					return 0;
				}

				Sleep (200);
				
			//Read the input buffer and discard till there are no characters to read.
				char szBuff[5] = {0};
				DWORD dwBytesRead = 0;
				if(!ReadFile(hSerial, szBuff, 4, &dwBytesRead, NULL))
				{
					std::cerr << "error occurred reading the input buffer";				
					Sleep(2000);
					return 0;
				}
		


			//Test Setup Questions
			


			//Quesetion 1, Tension or Compression Test?
				int test;
				std::cout << "\n\nDo you want a tension or compression test (tension = 1, compression = 2)?: "; 
				std::cin >> test;			
				if (test!=1 && test!=2)
				{
					std::cerr<< "Invalid Entry!";
					Sleep(2000);
					return 0;
				}

			//Question 2, Test Speed	
				double speed;
				std::cout << "\nEnter the desired test speed (mm/min linear travel): ";
				std::cin >> speed;
			
			//convert specified mm/min speed into motor rotation speed for open loop command
				double rot_speed,alpha,beta; 
				if (test == 1) // tension test, speeds 1024-2046 (motor rotates CW)	
				{
					beta = -39.7760880746;
					alpha = 0.0362830709;
				}		
				else if (test == 2) //compression test, speeds 1-1023 (motor rotates CCW)		
				{		
					beta = -1.4641653467;
					alpha = 0.0348805410;
				}
				rot_speed = (speed - beta)/alpha;

				if (test == 1) // if this is a tension test
				{
					if(rot_speed>2045) //bound the speed to tension speeds
						rot_speed=2045;
					else if(rot_speed<1025)
						rot_speed = 1025;
				}
				else if (test == 2) //if this is a compression test
				{
					if(rot_speed>1022)
						rot_speed=1022;
					else if(rot_speed<2)
						rot_speed=2;
				}

				double rot_speed1 = rot_speed; //save a copy of the initial speed

			//Question 3, how to stop the test
				int stopType;
				std::cout << "\nStop the test on sample break (1) or after set time (2)?: ";
				std::cin >> stopType;
				if (stopType !=1 && stopType!=2)
				{
					std::cerr << "Invalid Entry!";
					Sleep(2000);
					return 0;		
				}

				double testTimer;
				if (stopType == 2)
				{					
					std::cout << "\nGive test timeout length (minutes): ";
					std::cin >> testTimer;
				}


			//Name your data file
					std::wstring fileName;
					std::wstring fileName2;
					std::wstring extension (L".txt");
					std::cout << "\nGive a name for your data file (do not include file extension): ";
					std::wcin >> fileName;
					fileName2 = fileName+extension;					
					wchar_t* fileName3 = const_cast<wchar_t*>( fileName2.c_str() );
				

			//Open the file for writing data
				std::ofstream newfile;
				newfile.open (fileName3); //open the data file
				if(!newfile) //if the file does not open
				{
					std::cout << "\n Error: data file could not be opened!" << std::endl;
					Sleep(2000);
					return 0;
				}
				

			//Write a heading to the file
				newfile << "freeLoader Data file for crosshead " << motorNumber << " : ";	
				if (test==1)	
					newfile << "Tension Test: ";		
				else		
					newfile << "Compression Test: ";		
				newfile << speed << " mm/min \n";
				
				if(stopType==1)
					newfile << "Test stopped on sample break \n";
				else
					newfile << "Test stopped after " << testTimer << " minutes \n";


				newfile << std::setw(12) << "\nTime(ms)";
				newfile << std::setw(8) << "Angle";		
				newfile << std::setw(12) << "Load (mlb)" << std::endl;	
				

			//Start the test
				int start = 0;	
				while (start != 1)
				{
					std::cout << "\n\n\n     Configuration complete. Enter 1 and the test will begin: ";
					std::cin >> start;
				}


			//Run the test
				std::cout << "\n\n\n    **Testing in progress.  Please wait for the test to complete.** \n";		
				std::cout << "\n\n\n      -- In an emergency, press any key to terminate the test -- \n" << std::endl;


			//Read motor encoders				
				int angle1 = read_encoder(ftHandleDYNA, motorNumber);	
				Sleep(100);
			
			//Initialize counters and gains for feedback control of speed
				int counter=0; //for smoothing overshoot in feedback
				int counter2=0;	// for buffer backup error
				int counter3=0; // for stuck in deadzone error
				double angle_prev,time_prev,angleTravel,timeTravel,nowSpeed,Verr_prev,dVdt,rot_speed_PID,speedErr ;
				double angle_now=0;
				double time_now=0;
				double Verr=0;
				double Verr_tot=0;		
				int data_time;		
				int load1 = 0;
	
			//GAINS!!!
				double Kp=14.09;
				double Ki=12.3;
				double Kd=0;

			//Initialize the timer
				int time_start = milliseconds(); //from tic_toc.h



			//Write Data to the file				
				while (1) 
				{

					//check for e-stop exit
					if( _kbhit() > 0 )
					{
						std::cout << "\n\n   Test Terminated By User";
						break;	
					}

					//check for load cell overload (5kN max is ~ 1120000 mlbs)
					if(load1 >= 1120000 || load1 <= -1120000)
					{
						std::cout << "\n\n   Force Overload!";
						break;
					}

					//check for sample break
					if(stopType==1)
					{
						if(milliseconds()-time_start > 15000 && load1 <100)
							// a break is a load less than .1 lbs occuring after 15 s of testing
						{
							std::cout << "\n\n   Specimen Break Detected";
							break;
						}
								
					}
					else // stop after time
					{
						if(milliseconds()-time_start >= testTimer*60*1000)
						{
							std::cout << "\n\n   Test Timed Out";
							break;
						}
					}
					

					//Spin the motor			
					motor_spin(ftHandleDYNA, motorNumber, rot_speed);			
					Sleep(30);

					//read current test time
					data_time = milliseconds()-time_start;			

					//Write O0W1<CR> to the port (this sends a bad load but prepares the good load for sending)
					if(!WriteFile(hSerial, "O0W1\r", 5, &dwBytesWrite, NULL))
						std::cerr << "\n\n error occurred writing O0W1<CR> to port";

					//Read motor encoder
					angle1 = read_encoder(ftHandleDYNA, motorNumber);			
			
					//Read load cell input buffer and discard
					char SzBuff[13] = {0};
					dwBytesRead = 0;
					if(!ReadFile(hSerial, SzBuff, 12, &dwBytesRead, NULL))
						std::cerr << "error occurred reading the load cell input buffer";	
			
					//Write O0W1<CR> to the port (this sends the good load and prepares a the bad one for next loop)
					if(!WriteFile(hSerial, "O0W1\r", 5, &dwBytesWrite, NULL))
						std::cerr << "\n\n error occurred writing O0W1<CR>  to port";

					//Read load cell input buffer and keep it
					char SZBuff[13] = {0};	
					dwBytesRead = 0;
					if(!ReadFile(hSerial, SZBuff, 12, &dwBytesRead, NULL))
						std::cerr << "error occurred reading the load cell input buffer";
					
					//Convert load cell buffer to int
					load1 = atof(SZBuff);
					
					//write test time to file
					newfile << "\n" << std::fixed << std::setprecision(0) << std::setw(12) << data_time; 
					
					//write angle to file
					newfile << std::setw(8) << angle1;
					
					//write load to file
					newfile << std::setw(12) << load1;

					//watch for 65532 error - occurs when motor buffer is backed up
					if (angle1 == 65532)
						counter2++;
					if (counter2>2)
					{
						std::cerr << "\n\n Exiting! 65532 Error!";
						break;
					}			
					

					//check for stuck in the deadzone	
					if (angle1<5 || angle1>1005)
						counter3++;
					else 
						counter3=0;
					if (counter3>10)
						rot_speed=rot_speed+50;


					//feedback angle to adjust speed	
					angle_prev = angle_now;
					angle_now = angle1;
					time_prev = time_now;
					time_now = data_time;

					if (angle_now<1005 && angle_now>5 && angle_prev<1005 && angle_prev>5) 
					{//if we are within the encoder's usable range find the speed
						angleTravel = abs(angle_now-angle_prev);
						timeTravel = (time_now-time_prev)*.001/60; //converted to minutes
						nowSpeed = angleTravel/timeTravel/2418.9; //converted to mm/min
			
						//Find errors for PID control
						speedErr = speed-nowSpeed; //speed is the desired speed of the user
						Verr_prev = Verr;
						Verr = speedErr;
						Verr_tot = Verr_tot+Verr;
						dVdt = (Verr-Verr_prev)/timeTravel;
					
						counter++;
						if (counter>3) //run in open loop until 3rd iteration to smooth initial overshoot
						{
							rot_speed_PID =(Kp*Verr + Ki*Verr_tot + Kd*dVdt);
							
							rot_speed = rot_speed1 + rot_speed_PID;

							if (test == 1) // if this is a tension test
							{
								if(rot_speed>2045) //bound the speed to tension speeds
									rot_speed=2045;
								else if(rot_speed<1025)
									rot_speed = 1025;
							}
							else if (test == 2) //if this is a compression test
							{
								if(rot_speed>1022)
									rot_speed=1022;
								else if(rot_speed<2)
									rot_speed=2;
							}
						}
					}
				}//END OF TEST

			
			//shut off the motor
				motor_spin(ftHandleDYNA, motorNumber, 0);

			//Write end of file
				newfile << "\n\neof";

			// Close the data file
				newfile.close(); 


			//Close USB2Dynamixel
				ftStatus = FT_Close(ftHandleDYNA); // close the motor		
				if (ftStatus != FT_OK) 		
					std::cerr << "\n\nUSB2Dynamixel failed to close";


			//display "press any key..." to close the console
				std::cout << "\n";
				system("pause"); 

			return 0;
			}
}